% This program describes one very simple world, but it is a milestone for AI development!

?-
  start_random_from(2), % make the random deterministic
  G_Pause := 0,
  G_One_Step := 0,
  G_FastSpeed := 0,
  G_Last_Speed := 0,
  array(links, 8*8, 0),
  array(states, 8, 0),
  array(buf_views, 50, 0),
  array(buf_actions, 50, 0),
  array(buf_events, 50, 0),
  array(events, 4*10, max_integer),
  init_model,
  font(10,25,"Courier"),
  window(title("Simple World for AI"), size(1200, 800), pos(100, 100), paint_indirectly).

init_model:-
  G_AllEvents := 3+random(2), % from 1 to 4
  G_Circuits := 1+random(2), % increase for more complicated models
  G_Time := 0,
  G_Time_last := 0,
  G_Active := 0,
  G_Active_last1 := 0,
  G_Active_last2 := 0,
  G_Events1 := 0,
  G_Events2 := 0,
  clean_buffers,
  init_events(0),
  init_Links.

time_func(end) :-
  check_pause,
  G_Time := G_Time + 1,
  set_window_text,
%                 observation given by the world
  Trace := states(G_Active),
  calc_observation(Trace, Observation),
  buf_views(G_Time mod 50):= Observation,
%                 first transition
  events_are_true(Events1, 1),
  choose_new_state(New1, Events1),
  (New1 =\= G_Active ->
    G_Time_last := G_Time,
    G_Active_last1 := G_Active,
    G_Active_last2 := New1,
    G_Events1 := links(8*G_Active+New1)/\Events1,
    G_Events2 := 0,
    G_Active := New1,
    buf_events(G_Time mod 50):= G_Events1
  else
    buf_events(G_Time mod 50):= 0
  ),
%                 action done by the agent
  buf_actions(G_Time mod 50):= random(2),
%                 second transition
  events_are_true(Events2, 0),
  choose_new_state(New2, Events2),
  (New2 =\= G_Active ->
    (G_Time =\= G_Time_last ->
      G_Active_last1 := G_Active,
      G_Active_last2 := G_Active,
      G_Events1 := 0
    ),
    G_Time_last := G_Time,
    G_Events2 := links(8*G_Active+New2)/\Events2,
    G_Active := New2,
    buf_events(G_Time mod 50):= G_Events1 \/ G_Events2
  ),
  update_window(_).

choose_new_state(New, Events):-
  New := G_Active,
  for(I, 0, 7),
    links(8*G_Active+I)/\Events =\= 0,
    (New =:= G_Active; random(100) < 33 -> New := I),
    fail.
choose_new_state(New, Events).

set_window_text :-
  set_text("Simple World for AI, step="+G_Time+(G_Pause =:= 1 -> (G_One_Step > 1 -> " (One Transition)" else " (Pause)") else ""), _).

check_pause:-
  (G_Pause =:= 1 ->
    (G_One_Step =:= 1 ->
      G_One_Step := 0
    else
      G_One_Step > 1,
      (G_One_Step =:= 3 ->
        G_One_Step := 2
      else
        G_Time =\= G_Time_last
      )
    )
  ).

menu_time(init) :-
  menu(normal, action(menu_pause), text("&Pause\tP") ),
  menu(separator),
  menu(normal, action(menu_one_step), text("&One Step\tO") ),
  menu(separator),
  menu(normal, action(menu_one_transition), text("&One Transition\tI") ),
  menu(separator),
  menu(normal, action(menu_slow_motion), text("&Slow Motion\tS") ),
  menu(checked, action(menu_normal_speed), text("&Normal Speed") ),
  menu(normal, action(menu_fast_speed), text("&Fast Speed\tF") ).

menu_change_the_model(press) :-
  init_model,
  set_window_text,
  update_window(_).

menu_fast_speed(press) :-
  G_FastSpeed:=(G_FastSpeed=:=1 -> 0 else 1),
  change_speed.
menu_normal_speed(press) :-
  (G_FastSpeed=\=0->
    G_FastSpeed:=0,
    change_speed
  ).
menu_slow_motion(press) :-
  G_FastSpeed:=(G_FastSpeed=:= -1 -> 0 else -1),
  change_speed.

change_speed:-
  modify_menu(G_Menu_Time, 8+G_Last_Speed, normal, _ ),
  modify_menu(G_Menu_Time, 8+G_FastSpeed, checked, _ ),
  G_Last_Speed := G_FastSpeed,
  Interval := (G_FastSpeed =:= 1 -> 0.01 else (G_FastSpeed =:= 0 -> 0.5 else 2)),
  kill_timer(_, G_Timer),
  G_Timer := set_timer(_, Interval, time_func).

menu_pause(press) :-
  G_Pause := 1 - G_Pause,
  modify_menu(G_Menu_Time, 1, (G_Pause =:= 1 -> checked else normal), _ ),
  G_One_Step := 0,
  set_window_text.

menu_one_step(press) :-
  (G_Pause =:= 1 ->
    G_One_Step := 1
  ),
  time_func(end).

menu_one_transition(press) :-
  G_Pause := 1,
  modify_menu(G_Menu_Time, 1, checked, _ ),
  G_One_Step := 3,
  time_func(end).

menu_help(press):-
  message("Simple World for AI",
  "This is a simple world for AI, but it is a milestone for AI development!

In this world there is only one event-driven model. Also, the trace is only one. You see the model, but the agent has to find this model only on the base of the observations. The only clue for the agent is the trace. Here we test the model by random actions, but the real agent will be able to choose his actions.

To see more comlicated world go to Simle Plus.pro.",i).

win_func(init) :-
  menu(normal, action(menu_change_the_model), text("&Change the model") ),
  G_Menu_Time := menu( pop_up, action(menu_time), text("&Time") ),
  menu( right, action(menu_help), text("&Help") ),
  G_Timer := set_timer(_, 0.5, time_func).

win_func(key_down(80, Rep)):- % P
  menu_pause(press).
win_func(key_down(79, Rep)):- % O
  menu_one_step(press).
win_func(key_down(73, Rep)):- % I
  menu_one_transition(press).
win_func(key_down(83, Rep)):- % S
  menu_slow_motion(press). 
win_func(key_down(70, Rep)):- % F
  menu_fast_speed(press). 


win_func(paint):-
  show_line("views   ", 0, buf_views).
win_func(paint):-
  show_line("actions ", 40, buf_actions).
win_func(paint):-
  show_line("events  ", 80, buf_events).

win_func(paint):-
  for(Event, 0, G_AllEvents-1),
    makeColor(Event, Color),
	(2**Event/\buf_events(G_Time mod 50) =\= 0->
      pen(10, Color)
    else
      pen(3, Color)
    ),
    line(600, 50+Event*80, 1000, 50+Event*80),
    line(970, 60+Event*80, 1000, 50+Event*80, 970, 40+Event*80),
    pen(3, Color),
    color_text(_, rgb(255,0,0)),
    Star := (event_is_true(Event, 0)-> "*" else " "),
    text_out(Star+2**Event+". ", pos(600,70+Event*80)),
    color_text(_, rgb(0,0,0)),
    print_event(Event, 0),
    for(I, 0, 7),
	  for(J, 0, 7),
	    links(8*I+J)/\2**Event =\= 0,
		R := (G_Active=:=J -> 40 else 25),
		arrow(I, J, R, Event),
		fail.

win_func(paint):-
  State:=states(G_Active),
  State > 0,
  makeColor(State, Color),
  pen(10, Color),
  line(600, 450, 1000, 450),
  color_text(_, rgb(255,0,0)),
  text_out("The trace", pos(600,410)),
  text_out("*1. ", pos(600,470)),
  color_text(_, rgb(0,0,0)),
  text_out("=> "),
  color_text(_, rgb(0,0,255)),
  text_out("v0=1"),
  fail.

win_func(paint):-
  pen(3, rgb(0,0,0)),
  for(I, 0, 7),
    calcCoords(X,Y,I),
    State:=states(I),
    State > 0,
    makeColor(State, Color),
    brush(Color),
    (G_Active=:=I ->
      R := 40,
      (G_Time mod 2 =:= 1 -> pen(6, rgb(0,0,0))),
        ellipse(X-R,Y-R,X+R,Y+R),
        pen(3, rgb(0,0,0))
      else
        R := 25,
        ellipse(X-R,Y-R,X+R,Y+R)
      ),
      fail.

print_event(Event, Count):-
  Cell := events(10*Event+Count),
  (Cell =\= max_integer ->
    print_event(Event, Count+1),
    Atom := Cell//2,
    (Count=:=0 ->
      color_text(_, rgb(0,0,255))
    ),
    text_out((Atom mod 2 =:= 0 -> "a" else "v") + Atom//2 + "=" + Cell mod 2),
    (Count>0 ->
      text_out(" & ")
    )
  ).

event_is_true(Event, Count):-
  Cell := events(10*Event+Count),
  (Cell =\= max_integer ->
    Atom := Cell//2,
    Buffer := (Atom mod 2 =:= 0 -> buf_actions else buf_views),
    Buffer((G_Time-Atom//2) mod 50) =:= Cell mod 2,
    event_is_true(Event, Count+1)
  ).

events_are_true(Events, Views):-
  Events := 0,
  for(Event, 0, G_AllEvents-1),
    Cell := events(10*Event),
    (Cell =:= max_integer; (Cell//2) mod 2 =:= Views, event_is_true(Event, 0) -> Events := Events \/ 2**Event),
    fail.
events_are_true(Events, Views).

show_line(Title, Y, Buffer):-
  color_text(_, rgb(0,0,0)),
  text_out(" "+Title+" -> ", pos(0,600+Y)),
  for(N, 1, 50),
    (N mod 5 =:=0 -> color_text(_, rgb(255,0,0))),
    (N mod 5 =:=1 -> color_text(_, rgb(0,0,255))),
    Value := Buffer((N+G_Time) mod 50),
    (Value=:=0 ->
      text_out(".")
    else (Y=\=80 ->
      text_out("*")
    else
      text_out(hex_string(Value))
    )),
    fail.

op( 700, xfx, ':==').

arrow(I, J, R, Event):-
  calcCoords(X1,Y1,I),
  calcCoords(X2,Y2,J),
  D:= sqrt((X1-X2)**2+(Y1-Y2)**2),
  [A,B]:==[X1-X2,Y1-Y2]/D,
  Sign := (X1>X2; X1=X2, Y1<Y2 -> 1 else -1),
  [X1,Y1]:==[X1,Y1]+(Sign*10*(Event-1.5))*[B,-1*A],
  [X2,Y2]:==[X2,Y2]+(Sign*10*(Event-1.5))*[B,-1*A],

  [X3,Y3]:==[X2,Y2]+R*[A,B],
  [X4,Y4]:==[X2,Y2]+(R+30)*[A,B],
  [X4a,Y4a]:==[X4,Y4]+10*[-1*B,A],
  [X4b,Y4b]:==[X4,Y4]+10*[B,-1*A],
  (I=:=G_Active_last1, J=:=G_Active_last2, 2**Event/\G_Events1 =\= 0; I=:=G_Active_last2, J=:=G_Active, 2**Event/\G_Events2 =\= 0 ->
    makeColor(Event, Color),
    pen(10, Color),
    Bold := 1
  else
    Bold := 0
  ),
  line(X4a,Y4a,X3,Y3,X4b,Y4b),
  line(X1,Y1,X3,Y3),
  (Bold=:=1->
    pen(3, Color)
  ).

calcCoords(X,Y,I):-
  X := 300 - 200*cos(I*pi/4),
  Y := 300 - 200*sin(I*pi/4).

makeColor(0, rgb(0,0,0)).
makeColor(1, rgb(0,0,255)).
makeColor(2, rgb(255,0,0)).
makeColor(3, rgb(0,128,0)).
makeColor(4, rgb(255,255,0)).
makeColor(5, rgb(0,255,255)).

calc_observation(1, 0).
calc_observation(2, 1).
calc_observation(3, Observation):- Observation := true_value(random(4)=:=0).
calc_observation(4, Observation):- Observation := true_value(random(4)=\=0).
calc_observation(5, Observation):- Observation := true_value(random(2)=\=0).

clean_buffers:- for(I, 0, 8*8-1), links(I):=0, fail.
clean_buffers:- for(I, 0, 8-1), states(I):=0, fail.
clean_buffers:- for(I, 0, 50-1), buf_actions(I):=0, buf_views(I):=0, buf_events(I):=0, fail.
clean_buffers:- for(I, 0, 4*10-1), events(I):=max_integer, fail.
clean_buffers.

init_events(Event):-
  Event < G_AllEvents,
  init_one_event(Event),
  (not(exist_the_same(Event))->
    Event := Event + 1
  ),
  init_events(Event).
init_events(Event).

init_one_event(Event):-
  random(4)=\=0, % make the event "True"
  First := random(2),
  event_tail(Event, First, 0).
init_one_event(Event).

event_tail(Event, Last, Count):-
  events(10*Event+Count) := 2*Last + random(2),
  Count<3, random(3)=\=0,
  Next := Last + 1 + random(3),  % increase 3 to look further in the past
  event_tail(Event, Next, Count+1).
  
exist_the_same(This):-
  for(Event, 0, This - 1),
    the_same_event(This, Event, 0).

the_same_event(E1, E2, Count):-
  events(10*E1+Count) =:= events(10*E2+Count),
  (events(10*E1+Count) =:= max_integer; the_same_event(E1, E2, Count + 1)).

init_Links:-
  G_Used_States := 1,
  G_Used_Events := 0,
  for(I, 1, G_Circuits),
    circuit(0).
init_Links:-
  G_AllEvents := G_Used_Events.

circuit(State):-
  State2 := random(7),
  (State=:=State2; State2 >= G_Used_States ->
    (G_Used_States < 8 ->
      State2 := G_Used_States,
      G_Used_States := G_Used_States + 1
    else
      State2 := 7
    )
  ),
  Event := random(min(G_Used_Events+1, G_AllEvents)),
  (Event =:= G_Used_Events ->
    G_Used_Events := G_Used_Events+1
  ),
  links(8*State+State2):=links(8*State+State2)\/2**Event,
  (states(State)=:=0 ->
    states(State):=min(random(7)+1, 5),
    (states(State)=:=1, not(posible(Event, 0, 0))-> states(State) := 3),
    (states(State)=:=2, not(posible(Event, 1, 0))-> states(State) := 4)
  ),
  State2 =\= 0,
  circuit(State2).

posible(Event, Value, Count):-
  Cell := events(10*Event+Count),
  (Cell =\= max_integer ->
    Atom := Cell//2,
    (Atom mod 2 =:= 1 -> % views
      Cell mod 2 =:= Value
    ),
    posible(Event, Value, Count+1)
  ).

[] :== []/S. 
[H|T] :== [H1|T1]/S :- 
  H:= H1/S,
  T:== T1/S.

[] :== []+S*[]. 
[H|T] :== [H1|T1]+S*[H2|T2] :- 
  H:=H1+S*H2,
  T:==T1+S*T2.

